from multichain_explorer.src.validators.validator import ValidatorInterface
import re

class EthValidator(ValidatorInterface):

    blockRegex = r"^[0-9]{1,20}$"
    addressRegex = r"^0x[a-fA-F0-9]{40}$"
    transactionRegex = r"^0x[a-f0-9]{64}$"


    def is_block(self, text: str) -> bool:
        """Checks whether the text matches an ETH block number signature"""
        return bool( re.match(self.blockRegex, text) )

    def is_address(self, text: str) -> bool:
        """Checks whether the text matches an ETH address signature"""
        return bool( re.match(self.addressRegex, text) )

    def is_transaction(self, text: str) -> bool:
        """Checks whether the text matches an ETH block transaction signature"""        
        return bool( re.match(self.transactionRegex, text) )



from cryptos import *
from multichain_explorer.src.models.blockchains import Blockchains
from multichain_explorer.src.services.coinmarketcap_service import CoinMarketCapService
from multichain_explorer.src.validators.btc.btc_validator import BtcValidator
from multichain_explorer.src.validators.validator import ValidatorInterface
from multichain_explorer.src.models.provider_options import ProviderOptions
from multichain_explorer.src.providers.provider import ProviderInterface


class BtcProvider(ProviderInterface):
    
    validator: ValidatorInterface
    coinMarketCapService : CoinMarketCapService


    def __init__(self):
        self.provider = Bitcoin()
        self.validator = BtcValidator()
        self.coinMarketCapService : CoinMarketCapService()


    def get_summary(self):
        data = CoinMarketCapService().fetchData(Blockchains.BTC.value)
        cryptoData = data['data'][Blockchains.BTC.value]
        currency = "USD"

        summary = {
            "name" : cryptoData['name'],
            "price" : cryptoData['quote'][currency]['price'],
            "marketCap" : cryptoData['quote'][currency]['market_cap'],
            "totalSupply" : cryptoData['total_supply']
        }
        return summary


    def get_blocks(self, num_blocks = 10, options = ProviderOptions()):
        """ Returns a list of BTC block data, default number of blocks is 10 """

        latest_block_number = self.provider.current_block_height()
        latest_blocks = []

        for block_number in range(latest_block_number, 
                                latest_block_number - num_blocks, 
                                -1):
            block_data = self.get_block_by_id(block_number)
            latest_blocks.append(block_data)

        return latest_blocks


    def get_block_by_id(self, block_id = 'latest', options = ProviderOptions()):
        #If id is a number cast to int
        try:
            if block_id == 'latest':
                block_id = self.provider.current_block_height()
            else:
                block_id = int(block_id) #Cast to int
                
            block = self.provider.block_info(block_id)
            block_data = {
                "id" : block_id,
                "miner" : "-",
                "difficulty" : str( self.count_leading_zeroes(block['hash']) ) + " (target)",
                "timestamp": block['timestamp'],
            }

            if options.raw:
                #TODO - Pass real raw block data in the rawData field 
                #block_data["rawData"] = block
                block_data["rawData"] = {}

        except ValueError as err:
            # Log exception
            raise

        return block_data


    def count_leading_zeroes(text : str):
       #Count leading zeroes in a string
        match = re.match(r'^0+', text)
        return 0 if match is None else len(match.group())


    def get_transactions(self, num_tx = 10, options = ProviderOptions()):
        """Returns a list of transactions, default number of transactions is 10"""
        try:
            block_height = self.provider.current_block_height()
            #Get latest transactions
            block_transactions = self.provider.block_info(block_height)['tx_hashes']

            latest_transactions = []

            # Get the last num_tx transactions from block (default is 10)
            for transaction_id in block_transactions[-num_tx:]:
                #Convert id from binary hex to string
                transaction_data = self.get_transaction_by_id(transaction_id)
                latest_transactions.append(transaction_data)

        except ValueError as err:
            # Log exception
            raise
        return latest_transactions


    def get_transaction_by_id(self, tx_id = 'latest', options = ProviderOptions()):
        """Get an Algorand transaction by id"""

        if tx_id == 'latest':
            block_height = self.provider.current_block_height()
            #Get latest transaction id
            tx_id = self.provider.block_info(block_height)['tx_hashes'][-1]
        
        tx = self.provider.fetchtx(tx_id)

        transaction_data = {
            "id"         : tx_id,
            "from"       : tx['inputs'][0]['prev_out']['addr'],
            "to"         : tx['out'][0]['addr'],
            "value"      : tx['out'][0]['value'],  # value in satoshis
            "block"      : tx['block_index']
        }

        return transaction_data


    def get_address(self, address_id):
        """Get the info of an Algorand account"""

        try:
            history = self.provider.history(address_id)
            balance = history['final_balance'] # balance in satoshis

            return {
                "address" : address_id,
                "balance" : balance
            }
        except Exception as err:
            raise

    
    def search_resource(self, search_text):
        if self.validator.is_block(search_text):
            block = self.get_block_by_id(search_text)
            return { "type" : "block", "data" : block }
        if self.validator.is_address(search_text):
            address = self.get_address(search_text)
            return { "type" : "address", "data" : address }
        if self.validator.is_transaction(search_text):
            transaction = self.get_transaction_by_id(search_text)
            return { "type" : "transaction", "data" : transaction }
            
        raise


import abc

class ValidatorInterface(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'is_block') and 
                callable(subclass.is_block) and 
                hasattr(subclass, 'is_address') and 
                callable(subclass.is_address) or
                hasattr(subclass, 'is_transaction') and 
                callable(subclass.is_transaction) and  
                NotImplemented)

    @abc.abstractmethod
    def is_block(self, text: str) -> bool:
        """Checks whether the text matches a block number signature
        
        Args:
            text: text to check against the block number signature
        
        Returns:
            True if the text matches the block number signature, False otherwise
        """
        raise NotImplementedError

    @abc.abstractmethod
    def is_address(self, text: str) -> bool:
        """Checks whether the text matches an address signature
        
        Args:
            text: text to validate against address regex

        Returns:
            True if the text matches the address regex, False otherwise
        """
        raise NotImplementedError

    @abc.abstractmethod
    def is_transaction(self, text: str) -> bool:
        """Checks whether the text matches a transaction signature
        
        Args:
            text: text to validate against transaction regex

        Returns:
            True if the text matches the transaction regex, False otherwise
        """
        raise NotImplementedError


from dataclasses import dataclass
from typing import Optional

@dataclass
class Block:
    """Class that holds block data returned by the provider"""
    id: str
    miner: str
    difficulty: str
    timestamp: str
    raw_data: Optional[str] = "None"



from multichain_explorer.src.models.blockchains import Blockchains
from multichain_explorer.src.services.coinmarketcap_service import CoinMarketCapService
from multichain_explorer.src.validators.luna.luna_validator import LunaValidator
from multichain_explorer.src.validators.validator import ValidatorInterface
from multichain_explorer.src.models.provider_options import ProviderOptions
from multichain_explorer.src.providers.provider import ProviderInterface

from typing import List
from terra_sdk.client.lcd import LCDClient

class LunaProvider(ProviderInterface):
    
    provider = LCDClient(chain_id="columbus-5", url="https://lcd.terra.dev")
    validator: ValidatorInterface = LunaValidator()
    coinMarketCapService : CoinMarketCapService = CoinMarketCapService()


    def __init__(self):
        pass


    def get_summary(self):
        data = CoinMarketCapService().fetchData(Blockchains.LUNA.value)
        cryptoData = data['data'][Blockchains.LUNA.value]
        currency = "USD"

        summary = {
            "name" : cryptoData['name'],
            "price" : cryptoData['quote'][currency]['price'],
            "marketCap" : cryptoData['quote'][currency]['market_cap'],
            "totalSupply" : cryptoData['total_supply']
        }
        return summary


    def get_blocks(self, num_blocks = 10, options = ProviderOptions()):
        latest_block_number = int( self.provider.tendermint.block_info()['block']['header']['height'] )
        latest_blocks = []

        for block_number in range(latest_block_number, 
                                latest_block_number - num_blocks, 
                                -1):
            block_data = self.get_block_by_id(block_number)
            latest_blocks.append(block_data)
        
        return latest_blocks


    def get_block_by_id(self, block_id = 'latest', options = ProviderOptions()):
        try:
            if block_id == 'latest':
                block = self.provider.tendermint.block_info()
            else:
                block_id = int(block_id) #Cast to int
                block = self.provider.tendermint.block_info(block_id)

            block_data = {
                "id" :block_id,
                "miner" : block['block']['header']['proposer_address'],
                "difficulty" : "-",
                "timestamp": block['block']['header']['time']
            }
            
            if options.raw:
                block_data["rawData"] = {}

        except Exception as e:
            raise


    def get_transactions(self, num_tx = 10, options = ProviderOptions()):
        """Returns a list of transactions, default number of transactions is 10"""
        try:
            block_transactions = self.provider.tx.tx_infos_by_height()
            latest_transactions = []

            # Get the last num_tx transactions from block (default is 10)
            for transaction in block_transactions[-num_tx:]:
                transaction_data = self.extractTxData(transaction, options)
                latest_transactions.append(transaction_data)

        except ValueError as err:
            # Log exception
            raise
        return latest_transactions


    def get_transaction_by_id(self, tx_id = 'latest', options = ProviderOptions()):
        """Get a Terra transaction by id"""
        if tx_id == 'latest':
            tx_id = self.provider.tx.tx_infos_by_height()[-1].txhash

        try:
            tx = self.provider.tx.tx_info(tx_id)
            transaction_data = self.extractTxData(tx, options)
            return transaction_data
        except Exception as err:
            raise


    def get_address(self, address_id):
        try:
            address = self.provider.bank.balance(address_id)
            address_info = {
                "address"   : address_id,
                "balance"   : self._print_balance(address[0].to_data())
            }

            return address_info
        except Exception as err:
            raise

    
    def search_resource(self, search_text):
        if self.validator.is_block(search_text):
            block = self.get_block_by_id(search_text)
            return { "type" : "block", "data" : block }
        if self.validator.is_address(search_text):
            address = self.get_address(search_text)
            return { "type" : "address", "data" : address }
        if self.validator.is_transaction(search_text):
            transaction = self.get_transaction_by_id(search_text)
            return { "type" : "transaction", "data" : transaction }
            
        raise


    def extractTxData(self, tx, options = ProviderOptions()) -> dict:
        """Extracts data from a transaction"""
        
        # Get the transaction data based on transaction type
        from_address = ""
        to_address = ""  
        value = ""

        if type( tx.tx.body.messages[0] ).__name__ == 'MsgSend':
            from_address = tx.tx.body.messages[0].from_address
            to_address = tx.tx.body.messages[0].to_address
            value = self._print_coins( tx.tx.body.messages[0].amount.to_data() )
        elif type( tx.tx.body.messages[0] ).__name__ == 'MsgExecuteContract':
            from_address = tx.tx.body.messages[0].sender
            to_address = tx.tx.body.messages[0].contract  
            value = self._print_coins( tx.tx.body.messages[0].coins.to_data() )
        elif type( tx.tx.body.messages[0] ).__name__ == 'MsgMultiSend':
            from_address = ""
            to_address = ""  
            value = ""
        elif type( tx.tx.body.messages[0] ).__name__ == 'MsgWithdrawDelegationReward':
            from_address = ""
            to_address = ""  
            value = ""
        elif type( tx.tx.body.messages[0] ).__name__ == 'MsgWithdrawValidatorCommission':
            from_address = ""
            to_address = ""  
            value = ""
        elif type( tx.tx.body.messages[0] ).__name__ == 'MsgAggregateExchangeRateVote':
            from_address = ""
            to_address = ""  
            value = ""

        tx_data = {
                "id"    : tx.txhash,
                "from"  : from_address,
                "to"    : to_address,
                "value" : value,
                "block" : tx.height
            }
        
        if options.raw:
            #TODO - Add raw data
            tx_data["rawData"] = {}

        return tx_data


    def _print_coins(self, coins: List[dict]) -> str:
        """Returns a string with the formatted coins ('denom: amount')
        
        Args:
            coins (List[dict]): List of coins

        returns:
            str: Formatted coins
        """

        if len(coins) == 0:
            return "0"
        else:
            format_coins = lambda coin: f"{coin['amount']} ({coin['denom']})"
            formatted_coins = map(format_coins, coins)
            return ", ".join(formatted_coins)


    def _print_balance(self, balances: List[dict]) -> str:
        """Returns a string with the formatted balances ('denom: amount')
        
        Args:
            balances (List[dict]): List of balances

        returns:
            str: Formatted balances
        """

        if len(balances) == 0:
            return "0"
        else:
            format_balance = lambda balance: f"{balance['amount']} ({balance['denom']})"
            formatted_balances = map(format_balance, balances)
            return ", ".join(formatted_balances)






from multichain_explorer.src.models.blockchains import Blockchains
from multichain_explorer.src.services.coinmarketcap_service import CoinMarketCapService
from multichain_explorer.src.validators.ada.ada_validator import AdaValidator
from multichain_explorer.src.validators.validator import ValidatorInterface
from multichain_explorer.src.models.provider_options import ProviderOptions
from multichain_explorer.src.providers.provider import ProviderInterface
from blockfrost import BlockFrostApi, ApiError

class AdaProvider(ProviderInterface):
    
    validator: ValidatorInterface
    coinMarketCapService : CoinMarketCapService


    def __init__(self):
        self.provider = BlockFrostApi(project_id='mainnetugnPOWiXzieqDhohU5S1UWUsyvaXHLCz')
        self.validator = AdaValidator()
        self.coinMarketCapService : CoinMarketCapService()


    def get_summary(self):
        data = CoinMarketCapService().fetchData(Blockchains.ADA.value)
        cryptoData = data['data'][Blockchains.ADA.value]
        currency = "USD"

        summary = {
            "name" : cryptoData['name'],
            "price" : cryptoData['quote'][currency]['price'],
            "marketCap" : cryptoData['quote'][currency]['market_cap'],
            "totalSupply" : cryptoData['total_supply']
        }
        return summary


    def get_blocks(self, num_blocks = 10, options = ProviderOptions()):
        latest_block_number = self.get_block_by_id('latest').height
        latest_blocks = []

        for block_number in range(latest_block_number, 
                                latest_block_number - num_blocks, 
                                -1):
            block_data = self.get_block_by_id(block_number)
            latest_blocks.append(block_data)
        
        return latest_blocks


    def get_block_by_id(self, block_id = 'latest', options = ProviderOptions()):
        try:
            if block_id == 'latest':
                block = self.provider.block_latest()
            else:
                block = self.provider.block(block_id)

            block_data = {
                "id" : block.height,
                "miner" : block.slot_leader,
                "difficulty" : "-",
                "timestamp" : block.time,
                "block_hash" : block.hash
            }

            if options.raw:
                    block_data["rawData"] = {}
            
            return block_data
        except ApiError as e:
            print(e)

        return block_data


    def get_transactions(self, num_tx = 10, options = ProviderOptions()):
        """Returns a list of transactions, default number of transactions is 10"""
        try:
            block_transactions = self.provider.block_latest_transactions()
            latest_transactions = []

            # Get the last num_tx transactions from block (default is 10)
            for transaction_id in block_transactions[-num_tx:]:
                #Convert id from binary hex to string
                transaction_data = self.get_transaction_by_id(transaction_id)
                latest_transactions.append(transaction_data)

        except ValueError as err:
            # Log exception
            raise
        return latest_transactions



    def get_transaction_by_id(self, tx_id = 'latest', options = ProviderOptions()):
        """Get a Cardano transaction by id"""

        if tx_id == 'latest':
            tx_id = self.provider.block_latest_transactions()[0]
        
        transaction = self.provider.transaction(tx_id)
        tx_utxos = self.provider.transaction_utxos(tx_id)

        transaction_data = {
            "id"    : transaction.hash, #tx_id,
            "from"  : tx_utxos.inputs[0].address, 
            "to"    : tx_utxos.outputs[0].address,
            "value" : tx_utxos.outputs[0].amount[0].quantity, 
            "block" : transaction.block_height,
            "block_hash" : transaction.block
        }

        return transaction_data


    def get_address(self, address_id):
        try:
            address_info = self.provider.address(address_id)
            return {
                "address"   : address_id,
                "balance"   : address_info.amount[0].quantity
            }
        except Exception as err:
            raise

    
    def search_resource(self, search_text):
        if self.validator.is_block(search_text):
            block = self.get_block_by_id(search_text)
            return { "type" : "block", "data" : block }
        if self.validator.is_address(search_text):
            address = self.get_address(search_text)
            return { "type" : "address", "data" : address }
        if self.validator.is_transaction(search_text):
            transaction = self.get_transaction_by_id(search_text)
            return { "type" : "transaction", "data" : transaction }
            
        raise


from dataclasses import dataclass

@dataclass
class Summary:
    """Class that holds summary data of a blockchain returned by the provider"""
    name: str
    price: str
    market_cap: str
    total_supply: str


from multichain_explorer.src.services.fetch_service import FetchService

class CoinMarketCapService():
    """
    This class is used to fetch data from the CoinMarketCap API 
    """
    API_URL: str = 'https://pro-api.coinmarketcap.com/v1/cryptocurrency/quotes/latest'
    API_KEY: str = ''

    def __init__(self):
        self.currency = "USD"
        self.fetchService = FetchService()

    def fetchData(self, symbol: str) -> dict:
        """
        Fetch data from the CoinMarketCap API

        Args:
            symbol: the symbol of the cryptocurrency

        Returns:
            The data as a dict
        """
        try:
            params = self.get_params(symbol)
            headers = self.get_headers()

            #TODO - Use class member service
            # data = self.fetchService.fetch_json(self.API_URL, params, headers)
            data = FetchService().fetch_json(
                                        self.API_URL,
                                        params,
                                        headers
                                    )
            return data
        except Exception as e:
            print(e)
            raise


    def get_params(self, symbol: str) -> dict:
        """
        Get the parameters for the API call

        Args:
            symbol: the symbol of the cryptocurrency

        Returns:
            The parameters as a dict
        """
        parameters = {
            'symbol': symbol,
            'convert': self.currency
        }
        return parameters


    def get_headers(self) -> dict:
        """
        Get the headers for the API call

        Returns:
            The headers as a dict
        """
        headers = {
            'Accepts': 'application/json',
            'X-CMC_PRO_API_KEY': self.API_KEY
        }
        return headers




from requests import Request, Session
from requests.exceptions import ConnectionError, Timeout, TooManyRedirects
import json


class FetchService():

    def __init__(self):
        pass

    def fetch_json(self, endpoint: str, parameters, headers) -> dict: 
        """
        Fetch JSON data from an endpoint

        Args:
            endpoint: the endpoint to fetch from
            parameters: the parameters to pass to the endpoint
            headers: the headers to pass to the endpoint
        
        Returns:
            The JSON data as a dict

        Raises:
            HTTPException: if the connection fails
        """
        session = Session()
        session.headers.update(headers)

        try:
            response = session.get(endpoint, params=parameters)
            data = json.loads(response.text)
            return data
        except (ConnectionError, Timeout, TooManyRedirects) as e:
            print(e)
            raise
  



from multichain_explorer.src.models.blockchains import Blockchains
from multichain_explorer.src.services.coinmarketcap_service import CoinMarketCapService
from multichain_explorer.src.validators.algo.algo_validator import AlgoValidator
from multichain_explorer.src.validators.validator import ValidatorInterface
from multichain_explorer.src.models.provider_options import ProviderOptions
from multichain_explorer.src.providers.provider import ProviderInterface

from algosdk.v2client import algod
from algosdk.v2client import indexer


class AlgoProvider(ProviderInterface):

    algod_address = "https://mainnet-algorand.api.purestake.io/ps2"
    indexer_address = "https://mainnet-algorand.api.purestake.io/idx2"

    validator: ValidatorInterface
    coinMarketCapService : CoinMarketCapService


    def __init__(self):
        #TODO - Load algod_token from config file
        algod_token = "mAwxZgtD2M83O8Mb3953n3rON6PaVmqA19uVx3Mf"
        headers = { "X-API-Key": algod_token }

        self.provider = algod.AlgodClient(algod_token, self.algod_address, headers)
        self.provider.indexer = indexer.IndexerClient(algod_token, self.indexer_address, headers)
        
        self.validator = AlgoValidator()
        self.coinMarketCapService : CoinMarketCapService()


    def get_summary(self):
        data = CoinMarketCapService().fetchData(Blockchains.ALGO.value)
        cryptoData = data['data'][Blockchains.ALGO.value]
        currency = "USD"

        summary = {
            "name" : cryptoData['name'],
            "price" : cryptoData['quote'][currency]['price'],
            "marketCap" : cryptoData['quote'][currency]['market_cap'],
            "totalSupply" : cryptoData['total_supply']
        }
        return summary


    def get_blocks(self, num_blocks = 10, options = ProviderOptions()):
        latest_block_number = self.provider.status()['last-round']
        latest_blocks = []

        for block_number in range(latest_block_number, 
                                latest_block_number - num_blocks, 
                                -1):
            block_data = self.get_block_by_id(block_number)
            latest_blocks.append(block_data)
        
        return latest_blocks


    def get_block_by_id(self, block_id = 'latest', options = ProviderOptions()):
        try:
            if block_id == 'latest':
                status = self.provider.status()
                last_round = status.get("last-round")
                block = self.provider.block_info(last_round)
            else:
                block_id = int(block_id) #Cast to int
                block = self.provider.block_info(block_id)

            block_data = {
                "id" : block['block']['rnd'],
                "miner" : "-",
                "difficulty" : "-",
                "timestamp": block['block']['ts']
            }

            if options.raw:
                    block_data["rawData"] = {}

            return block_data
        except Exception as e:
            print(e)


    def get_transactions(self, num_tx = 10, options = ProviderOptions()):
        """Returns a list of transactions, default number of transactions is 10"""
        try:
            last_round = self.provider.status()['last-round']
            #Get latest transactions
            block_transactions = self.provider.indexer.search_transactions(block=last_round)['transactions']
            latest_transactions = []

            # Get the last num_tx transactions from block (default is 10)
            for transaction in block_transactions[-num_tx:]:
                #Convert id from binary hex to string
                transaction_data = self.get_transaction_by_id(transaction['id'])
                latest_transactions.append(transaction_data)

        except ValueError as err:
            # Log exception
            raise
        return latest_transactions


    def get_transaction_by_id(self, tx_id = 'latest', options = ProviderOptions()):
        """Get an Algorand transaction by id"""

        if tx_id == 'latest':
            last_round = self.provider.status()['last-round']
            #Get latest transaction id
            tx_id = self.provider.indexer.search_transactions(block=last_round)['transactions'][-1]['id']
        
        transaction = self.provider.indexer.transaction(tx_id)

        #Get receiver based on transaction type
        receiver = '-'
        value = 0
        tx_type = transaction['transaction']['tx-type']

        if tx_type == 'pay':
            tx_type_details = transaction['transaction']['payment-transaction']
            receiver = tx_type_details['receiver']
            value = tx_type_details['amount']
        elif tx_type == 'axfer':
            tx_type_details = transaction['transaction']['asset-transfer-transaction']
            receiver = tx_type_details['receiver']
            value = tx_type_details['amount']

        transaction_data = {
            "id"         : transaction['transaction']['id'],
            "from"       : transaction['transaction']['sender'],
            "to"         : receiver,
            "value"      : value, 
            "block"      : transaction['transaction']['confirmed-round'],
        }

        return transaction_data


    def get_address(self, address_id):
        """Get the info of an Algorand account"""
        try:
            account_info = self.provider.account_info(address_id)
            return {
                "address" : address_id,
                "balance" : account_info.get("amount")
            }
        except Exception as err:
            raise

    
    def search_resource(self, search_text):
        if self.validator.is_block(search_text):
            block = self.get_block_by_id(search_text)
            return { "type" : "block", "data" : block }
        if self.validator.is_address(search_text):
            address = self.get_address(search_text)
            return { "type" : "address", "data" : address }
        if self.validator.is_transaction(search_text):
            transaction = self.get_transaction_by_id(search_text)
            return { "type" : "transaction", "data" : transaction }
            
        raise


from enum import Enum

class Blockchains(Enum):
    ETH = "ETH"
    BTC = "BTC"
    ADA = "ADA"
    ALGO = "ALGO"
    LUNA = "LUNA"

    def get_available_blockchains():
        """
        Get all the blockchain id's

        Returns:
            The blockchains ids as a list
        """
        available_blockchains = [] 
        for blockchain in Blockchains:
            available_blockchains.append(blockchain)
        return available_blockchains


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['multichain_explorer',
 'multichain_explorer.src',
 'multichain_explorer.src.models',
 'multichain_explorer.src.models.provider_models',
 'multichain_explorer.src.providers',
 'multichain_explorer.src.providers.ada',
 'multichain_explorer.src.providers.algo',
 'multichain_explorer.src.providers.btc',
 'multichain_explorer.src.providers.eth',
 'multichain_explorer.src.providers.luna',
 'multichain_explorer.src.services',
 'multichain_explorer.src.validators',
 'multichain_explorer.src.validators.ada',
 'multichain_explorer.src.validators.algo',
 'multichain_explorer.src.validators.btc',
 'multichain_explorer.src.validators.eth',
 'multichain_explorer.src.validators.luna']

package_data = \
{'': ['*']}

install_requires = \
['blockfrost-python>=0.4.3,<0.5.0',
 'cryptos>=1.36,<2.0',
 'py-algorand-sdk>=1.11.0,<2.0.0',
 'terra-sdk>=2.0.5,<3.0.0',
 'web3>=5.28,<6.0']

entry_points = \
{'console_scripts': ['my-project = multichain_explorer.main:run']}

setup_kwargs = {
    'name': 'multichain-explorer',
    'version': '0.1.3',
    'description': 'A simple package to explore multiple blockchains in an homogeneous way',
    'long_description': None,
    'author': 'sdominguez894',
    'author_email': 'sdominguez894@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from multichain_explorer.src.validators.validator import ValidatorInterface
import re

class AlgoValidator(ValidatorInterface):

    blockRegex = r"^$"
    addressRegex = r"^$"
    transactionRegex = r"^$"


    def is_block(self, text: str) -> bool:
        """Checks whether the text matches an ALGO block number signature"""
        return bool( re.match(self.blockRegex, text) )

    def is_address(self, text: str) -> bool:
        """Checks whether the text matches an ALGO address signature"""
        return bool( re.match(self.addressRegex, text) )

    def is_transaction(self, text: str) -> bool:
        """Checks whether the text matches an ALGO block transaction signature"""        
        return bool( re.match(self.transactionRegex, text) )



from dataclasses import dataclass
from typing import Optional

@dataclass
class Transaction:
    """Class that holds transaction data returned by the provider"""
    id: str
    address_from: str
    address_to: str
    value: str
    block: str
    raw_data: Optional[str] = None
    



from multichain_explorer.src.providers.ada.ada_provider import AdaProvider
from multichain_explorer.src.providers.algo.algo_provider import AlgoProvider
from multichain_explorer.src.providers.btc.btc_provider import BtcProvider
from multichain_explorer.src.providers.eth.eth_provider import EthProvider
from multichain_explorer.src.providers.luna.luna_provider import LunaProvider
from multichain_explorer.src.providers.provider import ProviderInterface
from multichain_explorer.src.models.blockchains import Blockchains

class BlockchainProvider():

    def get_instance(blockchain_id: Blockchains) -> ProviderInterface:
        """
        Get the blockchain provider instance

        Args:
            blockchain_id: the blockchain id as an enum
        
        Returns:
            The blockchain provider instance wrapped in a ProviderInterface
        """
        match blockchain_id:
            case Blockchains.BTC:
                return BtcProvider()
            case Blockchains.ETH:
                return EthProvider()
            case Blockchains.ADA:
                return AdaProvider()
            case Blockchains.ALGO:
                return AlgoProvider()
            case Blockchains.LUNA:
                return LunaProvider()


import abc
from typing import List
from multichain_explorer.src.models.provider_options import ProviderOptions
from multichain_explorer.src.models.provider_models import Address, Block

class ProviderInterface(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'get_summary') and 
                callable(subclass.get_summary) and 
                hasattr(subclass, 'get_blocks') and 
                callable(subclass.get_blocks) and 
                hasattr(subclass, 'get_block_by_id') and 
                callable(subclass.get_block_by_id) and
                hasattr(subclass, 'get_transactions') and 
                callable(subclass.get_transactions) and 
                hasattr(subclass, 'get_transaction_by_id') and 
                callable(subclass.get_transaction_by_id) and 
                callable(subclass.get_address) and 
                hasattr(subclass, 'get_address') and 
                callable(subclass.search_resource) and 
                hasattr(subclass, 'search_resource') or 
                NotImplemented)

    @abc.abstractmethod
    def get_summary(self):
        """
        Get a summary of the blockchain

        Returns:
            A dict containing the summary data of the blockchain
            The dict has the following structure:
            {
                "name" : blockchain_name,
                "price" : token_price,
                "marketCap" : market_cap,
                "totalSupply" : total_supply
            }
                
        Raises:
            NotImplementedError if the method is not implemented
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_blocks(self, num_blocks: int, options: ProviderOptions):
        """ Get a list of blocks
        
        Args:
            num_blocks: number of blocks to return
            options: ProviderOptions object 

        Returns:
            The list of blocks
            
        Raises:
            NotImplementedError if the method is not implemented
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_block_by_id(self, block_id: str, options: ProviderOptions):
        """ Get a block by id
        
        Args:
            block_id: id of the block
            options: ProviderOptions object 

        Returns:
            A dict containing the block data
            The dict has the following structure:
            {
                "id"            : block_number,
                "miner"         : miner,
                "difficulty"    : difficulty,
                "timestamp"     : timestamp,
                "rawData"       : raw_data (optional)
            }
            
        Raises:
            NotImplementedError if the method is not implemented
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_transactions(self, num_tx: int, options: ProviderOptions):
        """Get a list of transactions
        
        Args:
            num_tx: number of transactions to return
            options: ProviderOptions object
        
        Returns:
            A list of transactions

        Raises:
            NotImplementedError if the method is not implemented
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_transaction_by_id(self, tx_id: str, options: ProviderOptions):
        """Get a transaction by id
        
        Args:
            tx_id: id of the transaction

        Returns:
            A dict containing the transaction data
            The dict has the following structure:
            {
                "id"    : tx_id,
                "from"  : transaction_from,
                "to"    : transaction_to,
                "value" : transaction_value,
                "block" : transaction_blockNumber
                "rawData" : transaction_rawData (optional)
            } 
        
        Raises:
            NotImplementedError if the method is not implemented
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_address(self, address_id):
        """Returns an address by its address identifier
        
        Args:
            address_id: address identifier

        Returns:
            A dict containing the address data
            The dict has the following structure:
            {
                "id"        : address_id,
                "balance"   : address_balance,
            }
        
        Raises:
            NotImplementedError if the method is not implemented
        """
        raise NotImplementedError

    @abc.abstractmethod
    def search_resource(self, search_text):
        """Search for a resource (block, address or transaction) by search_text
        
        Args:
            search_text: the text to search for
        
        Returns:
            A dict containing the resource (block, address or transaction) data
            The dict has the following structure:
            {
                "type"  : type ("block", "address" or "transaction"),
                "data"  : dict with block, address or transaction data
            }
        
        Raises:
            NotImplementedError if the method is not implemented
        """
        raise NotImplementedError


from web3   import Web3
from multichain_explorer.src.models.blockchains import Blockchains
from multichain_explorer.src.services.coinmarketcap_service import CoinMarketCapService
from multichain_explorer.src.validators.eth.eth_validator import EthValidator
from multichain_explorer.src.validators.validator import ValidatorInterface
from multichain_explorer.src.models.provider_options import ProviderOptions
from multichain_explorer.src.providers.provider import ProviderInterface


class EthProvider(ProviderInterface):
    
    INFURA_URL : str = ''
    validator: ValidatorInterface
    coinMarketCapService : CoinMarketCapService


    def __init__(self):
        self.provider = Web3(Web3.HTTPProvider(self.INFURA_URL))
        self.validator = EthValidator()
        self.coinMarketCapService : CoinMarketCapService()


    def get_summary(self):
        data = CoinMarketCapService().fetchData(Blockchains.ETH.value)
        cryptoData = data['data'][Blockchains.ETH.value]
        currency = "USD"

        summary = {
            "name" : cryptoData['name'],
            "price" : cryptoData['quote'][currency]['price'],
            "marketCap" : cryptoData['quote'][currency]['market_cap'],
            "totalSupply" : cryptoData['total_supply']
        }
        return summary


    def get_blocks(self, num_blocks = 10, options = ProviderOptions()):
        latest_blocks = []
        # Would be nice if calls were asynchronous
        for block_number in range(self.provider.eth.block_number, 
                                  self.provider.eth.block_number-num_blocks, 
                                  -1):
            block_data = self.get_block_by_id(block_number, options)
            latest_blocks.append(block_data)
        return latest_blocks


    def get_block_by_id(self, block_id = 'latest', options = ProviderOptions()):
        #If id is a number cast to int
        if type(block_id) == str and block_id.isnumeric():
            block_id = int(block_id)
        
        try:
            block = self.provider.eth.get_block(block_id)
            block_data = {
                "id" : block.number,
                "miner" : block.miner,
                "difficulty" : block.difficulty,
                "timestamp": block.timestamp,
            }

            if options.raw:
                block_data["rawData"] = self.provider.toJSON(block)

        except ValueError as err:
            # Log exception
            raise

        return block_data


    def get_transactions(self, num_tx = 10, options = ProviderOptions()):
        """Returns a list of transactions, default number of transactions is 10"""
        try:
            block = self.provider.eth.get_block("latest")
            block_transactions = block.transactions
            latest_transactions = []

            # Get the last num_tx transactions from block (default is 10)
            for transaction_id in block_transactions[-num_tx:]:
                #Convert id from binary hex to string
                transaction_data = self.get_transaction_by_id(transaction_id.hex(), options)
                latest_transactions.append(transaction_data)

        except ValueError as err:
            # Log exception
            raise
        return latest_transactions


    def get_transaction_by_id(self, tx_id = 'latest', options = ProviderOptions()):
        """Get an Ethereum transaction by id"""
        transaction = self.provider.eth.get_transaction(tx_id)
        transaction_data = {
            "id"    : tx_id,
            "from"  : transaction['from'],
            "to"    : transaction['to'],
            "value" : transaction['value'],
            "block" : transaction['blockNumber']
        }

        if options.raw:
            transaction_data["rawData"] = self.provider.toJSON(transaction)

        return transaction_data


    def get_address(self, address_id):
        try:
            balance = self.provider.eth.get_balance(address_id)
            balance = self.provider.fromWei(balance, 'ether')
            return {
                "address"   : address_id,
                "balance"   : balance
            }
        except Exception as err:
            #log error
            raise

    
    def search_resource(self, search_text):
        if self.validator.is_block(search_text):
            block = self.get_block_by_id(search_text)
            return { "type" : "block", "data" : block }
        if self.validator.is_address(search_text):
            address = self.get_address(search_text)
            return { "type" : "address", "data" : address }
        if self.validator.is_transaction(search_text):
            transaction = self.get_transaction_by_id(search_text)
            return { "type" : "transaction", "data" : transaction }
            
        raise